#!/usr/bin/env python
# License: GPLv3 Copyright: 2025, Kovid Goyal <kovid at kovidgoyal.net>

# Atexit that works even if the process crashes

import atexit
import json
import os
import sys
import time
from contextlib import suppress
from functools import wraps
from threading import RLock

_plat = sys.platform.lower()
iswindows = 'win32' in _plat or 'win64' in _plat

lock = RLock()
worker = None
RMTREE_ACTION = 'rmtree'
UNLINK_ACTION = 'unlink'


def thread_safe(f):

    @wraps(f)
    def wrapper(*a, **kw):
        with lock:
            return f(*a, **kw)
    return wrapper


@thread_safe
def remove_folder_atexit(path: str) -> None:
    _send_command(RMTREE_ACTION, os.path.abspath(path))


@thread_safe
def remove_file_atexit(path: str) -> None:
    _send_command(UNLINK_ACTION, os.path.abspath(path))


def unlink(path):
    with suppress(Exception):
        import os as oss
        if oss.path.exists(path):
            oss.remove(path)


def close_worker(worker):
    import subprocess
    worker.stdin.close()
    try:
        worker.wait(10)
    except subprocess.TimeoutExpired:
        worker.kill()
    worker.wait()


def ensure_worker():
    global worker
    if worker is None:
        from calibre.utils.ipc.simple_worker import start_pipe_worker
        worker = start_pipe_worker('from calibre.utils.safe_atexit import main; main()', stdout=None)
        atexit.register(close_worker, worker)
    return worker


def reset_after_fork():
    global worker
    atexit.unregister(close_worker)
    worker = None


def _send_command(action: str, payload: str) -> None:
    worker = ensure_worker()
    worker.stdin.write(json.dumps({'action': action, 'payload': payload}).encode('utf-8'))
    worker.stdin.write(os.linesep.encode())
    worker.stdin.flush()


if iswindows:
    def remove_dir(x):
        from calibre.utils.filenames import make_long_path_useable
        x = make_long_path_useable(x)
        import shutil
        import time
        for i in range(10):
            try:
                shutil.rmtree(x)
                return
            except Exception:
                import os  # noqa
                if os.path.exists(x):
                    # In case some other program has one of the files open.
                    time.sleep(0.1)
                else:
                    return
        with suppress(Exception):
            shutil.rmtree(x, ignore_errors=True)
else:
    def remove_dir(x):
        import shutil
        with suppress(Exception):
            shutil.rmtree(x, ignore_errors=True)


def main():
    if not iswindows:
        import signal
        signal.signal(signal.SIGINT, signal.SIG_IGN)
    ac_map = {RMTREE_ACTION: remove_dir, UNLINK_ACTION: unlink}
    for line in sys.stdin.buffer:
        if line:
            try:
                cmd = json.loads(line)
                atexit.register(ac_map[cmd['action']], cmd['payload'])
            except Exception:
                import traceback
                traceback.print_exc()


def main_for_test(do_forced_exit=False, check_tdir=False):
    if check_tdir:
        import tempfile

        from calibre.ptempfile import base_dir
        print(tempfile.gettempdir())
        print(base_dir())
        return
    tf = 'test-folder'
    os.mkdir(tf)
    open(os.path.join(tf, 'test-file'), 'w').close()
    remove_folder_atexit(tf)
    if do_forced_exit:
        os._exit(os.EX_OK)
    else:
        sys.stdin.read()


def find_tests():
    import tempfile
    import unittest

    from calibre.utils.ipc.simple_worker import start_pipe_worker

    class TestSafeAtexit(unittest.TestCase):

        def wait_for_empty(self, tdir, timeout=10):
            st = time.monotonic()
            while time.monotonic() - st < timeout:
                q = os.listdir(tdir)
                if not q:
                    break
                time.sleep(0.01)
            self.assertFalse(q)

        def test_safe_atexit(self):
            with tempfile.TemporaryDirectory() as tdir:
                self.assertFalse(os.listdir(tdir))
                p = start_pipe_worker('from calibre.utils.safe_atexit import main_for_test; main_for_test()', cwd=tdir)
                p.stdin.close()
                p.wait(10)
                self.wait_for_empty(tdir)
                p = start_pipe_worker('from calibre.utils.safe_atexit import main_for_test; main_for_test()', cwd=tdir)
                p.kill()
                p.wait(10)
                self.wait_for_empty(tdir)
                p = start_pipe_worker('from calibre.utils.safe_atexit import main_for_test; main_for_test(True)', cwd=tdir)
                p.wait(10)
                self.wait_for_empty(tdir)
                p = start_pipe_worker('from calibre.utils.safe_atexit import main_for_test; main_for_test(check_tdir=True)')
                tempfiledir, bdir = p.stdout.read().decode().splitlines()[:2]
                p.wait(10)
                self.assertEqual(bdir, tempfiledir)

    return unittest.defaultTestLoader.loadTestsFromTestCase(TestSafeAtexit)



#!/usr/bin/env python
# License: GPL v3 Copyright: 2020, Kovid Goyal <kovid at kovidgoyal.net>

import errno
import os
import socket
import sys
from contextlib import closing
from functools import partial
from itertools import count

from qt.core import QAbstractSocket, QLocalServer, pyqtSignal

from calibre.utils.ipc import gui_socket_address, socket_address


def unix_socket(timeout=10):
    ans = socket.socket(socket.AF_UNIX)
    ans.settimeout(timeout)
    return ans


class Listener(QLocalServer):

    message_received = pyqtSignal(object)

    def __init__(self, address=None, parent=None):
        QLocalServer.__init__(self, parent)
        self.address = address or gui_socket_address()
        self.uses_filesystem = self.address[0] not in '\0\\'
        self.setSocketOptions(QLocalServer.SocketOption.UserAccessOption)
        self.newConnection.connect(self.on_new_connection)
        self.connection_id = count()
        self.pending_messages = {}

    def start_listening(self):
        if self.address.startswith('\0'):
            s = unix_socket()
            s.bind(self.address)
            s.listen(16)
            if not self.listen(s.detach()):
                raise OSError(f'Could not start Listener for IPC at address @{self.address[1:]} with error: {self.errorString()}')
        else:
            if not self.listen(self.address):
                if self.serverError() == QAbstractSocket.SocketError.AddressInUseError and self.uses_filesystem:
                    self.removeServer(self.address)
                    if self.listen(self.address):
                        return
                code = self.serverError()
                if code == QAbstractSocket.SocketError.AddressInUseError:
                    raise OSError(errno.EADDRINUSE, os.strerror(errno.EADDRINUSE), self.address)
                raise OSError(f'Could not start Listener for IPC at address {self.address} with error: {self.errorString()}')

    def on_new_connection(self):
        while True:
            s = self.nextPendingConnection()
            if s is None:
                break
            cid = next(self.connection_id)
            self.pending_messages[cid] = b''
            s.readyRead.connect(partial(self.on_ready_read, cid, s))
            s.disconnected.connect(partial(self.on_disconnect, cid, s))

    def on_ready_read(self, connection_id, q_local_socket):
        num = q_local_socket.bytesAvailable()
        if num > 0:
            self.pending_messages[connection_id] += bytes(q_local_socket.readAll())

    def on_disconnect(self, connection_id, q_local_socket):
        self.on_ready_read(connection_id, q_local_socket)
        q_local_socket.close()
        q_local_socket.readyRead.disconnect()
        q_local_socket.disconnected.disconnect()
        q_local_socket.deleteLater()
        self.message_received.emit(self.pending_messages.pop(connection_id, b''))


def send_message_in_process(msg, address=None, timeout=5):
    address = address or gui_socket_address()
    if isinstance(msg, str):
        msg = msg.encode('utf-8')
    if address.startswith('\\'):
        with open(address, 'r+b') as f:
            f.write(msg)
    else:
        ps = unix_socket(timeout)
        ps.connect(address)
        ps.sendall(msg)
        ps.shutdown(socket.SHUT_RDWR)
        ps.close()


def send_message_in_worker(address, timeout):
    msg = sys.stdin.buffer.read()
    send_message_in_process(msg, address, timeout)


def send_message_via_worker(msg, address=None, timeout=5, wait_till_sent=True):
    # On Windows sending a message in a process that also is listening on the
    # same named pipe in a different thread deadlocks, so we do the actual sending in
    # a simple worker process
    from calibre.utils.ipc.simple_worker import start_pipe_worker
    if isinstance(msg, str):
        msg = msg.encode('utf-8')
    p = start_pipe_worker(f'from calibre.gui2.listener import *; send_message_in_worker({address!r}, {timeout!r})')
    with closing(p.stdin):
        p.stdin.write(msg)
    if wait_till_sent:
        return p.wait(timeout=timeout + 2) == 0


def listener_for_test(address):
    from qt.core import QCoreApplication, QTimer
    app = QCoreApplication([])
    s = Listener(address=address, parent=app)
    s.start_listening()
    def got_message(msg):
        if msg == b'quit':
            app.quit()
            return
        sys.stdout.buffer.write(msg)
        sys.stdout.buffer.write(os.linesep.encode())
        sys.stdout.buffer.flush()
    s.message_received.connect(got_message)
    QTimer.singleShot(0, lambda: print('started', flush=True))
    app.exec()


def find_tests():
    import unittest
    class TestIPC(unittest.TestCase):

        def test_listener_ipc(self):
            from calibre.utils.ipc.simple_worker import start_pipe_worker
            address = socket_address('test')
            server = start_pipe_worker(f'from calibre.gui2.listener import *; listener_for_test({address!r})')
            try:
                self.assertEqual(server.stdout.readline().rstrip(), b'started')
                for msg in 'one', 'two', 'three':
                    send_message_in_process(msg, address=address)
                    self.assertEqual(server.stdout.readline().rstrip(), msg.encode())
                self.assertTrue(send_message_via_worker('hello, world!', address=address))
                self.assertEqual(server.stdout.readline().rstrip(), b'hello, world!')
                msg = '123456789' * 8192 * 10
                send_message_in_process(msg, address=address)
                self.assertEqual(server.stdout.readline().rstrip(), msg.encode())
                send_message_in_process('quit', address=address)
                server.wait(2)
            finally:
                server.kill()
                server.wait()

    return unittest.defaultTestLoader.loadTestsFromTestCase(TestIPC)


def test():
    from qt.core import QApplication, QLabel, QTimer
    app = QApplication([])
    l = QLabel()
    l.setText('Waiting for message...')
    m = '123456789' * 8192 * 10

    def show_message(msg):
        q = msg.decode()
        if q != m:
            print(f'Received incorrect message of length: {len(q)} expecting length: {len(m)}', file=sys.stderr)
        else:
            print('Received msg correctly', file=sys.stderr)
        l.setText(q)
        QTimer.singleShot(1000, app.quit)

    def send():
        send_message_via_worker(m, wait_till_sent=False)

    QTimer.singleShot(1000, send)
    s = Listener(parent=l)
    s.start_listening()
    print('Listening at:', s.serverName(), s.isListening())
    s.message_received.connect(show_message)

    l.show()
    app.exec()


if __name__ == '__main__':
    test()



#!/usr/bin/env python

__license__   = 'GPL v3'
__copyright__ = '2025, Kovid Goyal <kovid@kovidgoyal.net>'
__docformat__ = 'restructuredtext en'


from threading import Thread

from qt.core import (
    QBrush,
    QColor,
    QColorDialog,
    QDialog,
    QHBoxLayout,
    QLabel,
    QPainter,
    QPixmap,
    QPushButton,
    QSize,
    QSizePolicy,
    Qt,
    QTabWidget,
    QWidget,
    pyqtSignal,
)

from calibre import human_readable
from calibre.gui2 import gprefs, open_local_file, question_dialog, resolve_grid_color
from calibre.gui2.library.alternate_views import CM_TO_INCH, auto_height
from calibre.gui2.preferences import LazyConfigWidgetBase
from calibre.gui2.preferences.look_feel_tabs.cover_grid_ui import Ui_Form
from calibre.gui2.widgets import BusyCursor
from calibre.startup import connect_lambda
from calibre.utils.icu import sort_key


class Background(QWidget):

    changed_signal = pyqtSignal()

    def __init__(self, parent):
        QWidget.__init__(self, parent)
        self.l = QHBoxLayout(self)
        self.l.setContentsMargins(0, 0, 0, 0)
        self.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)
        self.initialized = False
        self.changed_signal.connect(self.update_brush)

    def load_from_gprefs(self, use_defaults=False):
        self.bcol_dark = QColor(*resolve_grid_color(for_dark=True, use_defaults=use_defaults))
        self.bcol_light = QColor(*resolve_grid_color(for_dark=False, use_defaults=use_defaults))
        self.btex_dark = resolve_grid_color('texture', for_dark=True, use_defaults=use_defaults)
        self.btex_light = resolve_grid_color('texture', for_dark=False, use_defaults=use_defaults)
        self.update_brush()

    def lazy_initialize(self):
        if self.initialized:
            return
        self.initialized = True
        l = self.layout()
        text = (
            '<p style="text-align: center; color: {}"><b>{}</b><br>'
            '<a style="text-decoration: none" href="la://color.me">{}</a><br>'
            '<a style="text-decoration: none" href="la://texture.me">{}</a></p>')
        self.light_label = la = QLabel(text.format('black', _('Light'), _('Change color'), _('Change texture')))
        la.linkActivated.connect(self.light_link_activated)
        l.addWidget(la)
        self.dark_label = la = QLabel(text.format('white', _('Dark'), _('Change color'), _('Change texture')))
        la.linkActivated.connect(self.dark_link_activated)
        l.addWidget(la)
        self.load_from_gprefs()

    def change_color(self, light=False):
        which = _('light') if light else _('dark')
        col = QColorDialog.getColor(self.bcol_light if light else self.bcol_dark,
            self, _('Choose {} background color for the Cover grid').format(which))

        if col.isValid():
            if light:
                self.bcol_light = col
            else:
                self.bcol_dark = col
            btex = self.btex_light if light else self.btex_dark
            if btex:
                if question_dialog(
                    self, _('Remove background image?'),
                    _('There is currently a background image set, so the color'
                      ' you have chosen will not be visible. Remove the background image?')):
                    if light:
                        self.btex_light = None
                    else:
                        self.btex_dark = None
            self.changed_signal.emit()

    def change_texture(self, light=False):
        from calibre.gui2.preferences.texture_chooser import TextureChooser
        btex = self.btex_light if light else self.btex_dark
        d = TextureChooser(parent=self, initial=btex)
        if d.exec() == QDialog.DialogCode.Accepted:
            if light:
                self.btex_light = d.texture
            else:
                self.btex_dark = d.texture
            self.changed_signal.emit()

    def light_link_activated(self, url):
        if 'texture' in url:
            self.change_texture(light=True)
        else:
            self.change_color(light=True)

    def dark_link_activated(self, url):
        if 'texture' in url:
            self.change_texture(light=False)
        else:
            self.change_color(light=False)

    def commit(self):
        s = gprefs['cover_grid_background'].copy()
        s['light'] = tuple(self.bcol_light.getRgb())[:3]
        s['dark'] = tuple(self.bcol_dark.getRgb())[:3]
        s['light_texture'] = self.btex_light
        s['dark_texture'] = self.btex_dark
        gprefs['cover_grid_background'] = s

    def restore_defaults(self):
        self.load_from_gprefs(use_defaults=True)

    def update_brush(self):
        self.light_brush = QBrush(self.bcol_light)
        self.dark_brush = QBrush(self.bcol_dark)
        def dotex(path, brush):
            if path:
                from calibre.gui2.preferences.texture_chooser import texture_path
                path = texture_path(path)
                if path:
                    p = QPixmap(path)
                    try:
                        dpr = self.devicePixelRatioF()
                    except AttributeError:
                        dpr = self.devicePixelRatio()
                    p.setDevicePixelRatio(dpr)
                    brush.setTexture(p)
        dotex(self.btex_light, self.light_brush)
        dotex(self.btex_dark, self.dark_brush)
        self.update()

    def sizeHint(self):
        return QSize(200, 120)

    def paintEvent(self, ev):
        self.lazy_initialize()
        painter = QPainter(self)
        r = self.rect()
        light = r.adjusted(0, 0, -r.width()//2, 0)
        dark = r.adjusted(light.width(), 0, 0, 0)
        painter.fillRect(light, self.light_brush)
        painter.fillRect(dark, self.dark_brush)
        painter.end()
        super().paintEvent(ev)


class CoverGridTab(QTabWidget, LazyConfigWidgetBase, Ui_Form):

    changed_signal = pyqtSignal()
    restart_now = pyqtSignal()
    size_calculated = pyqtSignal(object)

    def __init__(self, parent=None):
        super().__init__(parent)

    def genesis(self, gui):
        self.gui = gui
        db = self.gui.library_view.model().db
        r = self.register

        r('cover_grid_width', gprefs)
        r('cover_grid_height', gprefs)
        r('cover_grid_cache_size_multiple', gprefs)
        r('cover_grid_disk_cache_size', gprefs)
        r('cover_grid_spacing', gprefs)
        r('cover_grid_show_title', gprefs)
        r('emblem_size', gprefs)
        r('emblem_position', gprefs, choices=[
            (_('Left'), 'left'), (_('Top'), 'top'), (_('Right'), 'right'), (_('Bottom'), 'bottom')])

        fm = db.field_metadata
        choices = sorted(((fm[k]['name'], k) for k in fm.displayable_field_keys() if fm[k]['name']),
                         key=lambda x:sort_key(x[0]))
        r('field_under_covers_in_grid', db.prefs, choices=choices)

        self.grid_rules.genesis(self.gui)
        self.grid_rules.changed_signal.connect(self.changed_signal)
        self.size_calculated.connect(self.update_cg_cache_size, type=Qt.ConnectionType.QueuedConnection)

        l = self.cg_background_box.layout()
        self.cg_bg_widget = w = Background(self)
        w.changed_signal.connect(self.changed_signal)
        l.addWidget(w, 0, 0, 3, 1)
        self.cover_grid_default_appearance_button = b = QPushButton(_('Restore default &appearance'), self)
        b.setSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
        l.addWidget(b, 0, 1)
        b.clicked.connect(self.restore_cover_grid_appearance)
        self.cover_grid_empty_cache.clicked.connect(self.empty_cache)
        self.cover_grid_open_cache.clicked.connect(self.open_cg_cache)
        connect_lambda(self.cover_grid_smaller_cover.clicked, self, lambda self: self.resize_cover(True))
        connect_lambda(self.cover_grid_larger_cover.clicked, self, lambda self: self.resize_cover(False))
        self.cover_grid_reset_size.clicked.connect(self.cg_reset_size)
        self.opt_cover_grid_disk_cache_size.setMinimum(self.gui.grid_view.thumbnail_cache.min_disk_cache)
        self.opt_cover_grid_disk_cache_size.setMaximum(self.gui.grid_view.thumbnail_cache.min_disk_cache * 100)
        self.opt_cover_grid_width.valueChanged.connect(self.update_aspect_ratio)
        self.opt_cover_grid_height.valueChanged.connect(self.update_aspect_ratio)

    def lazy_initialize(self):
        self.show_current_cache_usage()

        self.blockSignals(True)
        self.grid_rules.lazy_initialize()
        self.lazy_init_called = True
        self.blockSignals(False)
        self.cg_bg_widget.lazy_initialize()
        self.update_aspect_ratio()

    def show_current_cache_usage(self):
        t = Thread(target=self.calc_cache_size)
        t.daemon = True
        t.start()

    def calc_cache_size(self):
        self.size_calculated.emit(self.gui.grid_view.thumbnail_cache.current_size)

    @property
    def current_cover_size(self):
        cval = self.opt_cover_grid_height.value()
        wval = self.opt_cover_grid_width.value()
        if cval < 0.1:
            dpi = self.opt_cover_grid_height.logicalDpiY()
            cval = auto_height(self.opt_cover_grid_height) / dpi / CM_TO_INCH
        if wval < 0.1:
            wval = 0.75 * cval
        return wval, cval

    def update_aspect_ratio(self):
        width, height = self.current_cover_size
        ar = width / height
        self.cover_grid_aspect_ratio.setText(_('Current aspect ratio (width/height): %.2g') % ar)

    def resize_cover(self, smaller):
        wval, cval = self.current_cover_size
        ar = wval / cval
        delta = 0.2 * (-1 if smaller else 1)
        cval += delta
        cval = max(0, cval)
        self.opt_cover_grid_height.setValue(cval)
        self.opt_cover_grid_width.setValue(cval * ar)

    def cg_reset_size(self):
        self.opt_cover_grid_width.setValue(0)
        self.opt_cover_grid_height.setValue(0)

    def open_cg_cache(self):
        open_local_file(self.gui.grid_view.thumbnail_cache.location)

    def update_cg_cache_size(self, size):
        self.cover_grid_current_disk_cache.setText(
            _('Current space used: %s') % human_readable(size))

    def empty_cache(self):
        self.gui.grid_view.thumbnail_cache.empty()
        self.calc_cache_size()

    def restore_cover_grid_appearance(self):
        self.cg_bg_widget.restore_defaults()
        self.changed_signal.emit()

    def commit(self):
        with BusyCursor():
            self.grid_rules.commit()
            self.cg_bg_widget.commit()
        return LazyConfigWidgetBase.commit(self)

    def restore_defaults(self):
        LazyConfigWidgetBase.restore_defaults(self)
        self.grid_rules.restore_defaults()
        self.cg_bg_widget.restore_defaults()
        self.changed_signal.emit()

    def refresh_gui(self, gui):
        gui.library_view.refresh_grid()
        gui.grid_view.refresh_settings()
        gui.update_auto_scroll_timeout()


if __name__ == '__main__':
    from calibre.gui2 import Application
    from calibre.gui2.preferences import test_widget
    app = Application([])
    test_widget('Interface', 'Look & Feel', callback=lambda w: w.sections_view.setCurrentRow(1))
